# -*- coding: utf-8 -*-

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.forms.models import ModelForm, ModelFormMetaclass, ModelFormOptions
from django.forms.utils import ErrorDict
from django.template.loader import render_to_string
from django.utils.translation import gettext as _

from attachments.models import Attachment

import warnings


class BBCodeModelFormOptions(ModelFormOptions):
    def __init__(self, options=None):
        super(BBCodeModelFormOptions, self).__init__(options=options)
        self.bbcode_field = getattr(options, "bbcode_field", None)
        self.main_attach_target = getattr(options, "main_attach_target", None)


class BBCodeModelFormMetaclass(ModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        new_class = super(BBCodeModelFormMetaclass, mcs).__new__(mcs, name, bases, attrs)

        new_class._meta = BBCodeModelFormOptions(getattr(new_class, "Meta", None))

        return new_class


class BBCodeModelForm(ModelForm, metaclass=BBCodeModelFormMetaclass):
    """
    Special ModelForm that can be used for BBCodeFields with or without
    attachments. It provides a preview method, and ensure that attachments
    are correctly committed when the form is saved.
    """

    def __init__(self, *args, **kwargs):
        self._namespace = kwargs.pop("namespace", None)
        self.attachments_submitted = False
        self.attachments_errors = []

        super(BBCodeModelForm, self).__init__(*args, **kwargs)

        opts = self._meta
        if opts.bbcode_field is None:
            raise ValueError("BBCodeModelForm has no bbcode_field meta parameter.")

        self._bbcode_field_name = opts.bbcode_field
        self._bbcode_field = self.fields.get(self._bbcode_field_name)
        self._main_attach_target_name = opts.main_attach_target

        if self._bbcode_field is None:
            raise ValueError(f"BBCode field {repr(self._bbcode_field_name)} is not a valid field.")

        self._has_attachments = self._bbcode_field.can_have_attachments

        if self._has_attachments and self._namespace is None:
            raise TypeError("Missing argument 'namespace'")

        if not self._has_attachments:
            if self._namespace is not None:
                raise TypeError(
                    "Argument 'namespace' is not needed -- the "
                    f"BBCode field {repr(self._bbcode_field_name)} does not use attachments"
                )

            if self._main_attach_target_name is not None:
                raise TypeError(
                    "Argument 'main_attach_target' is not needed "
                    f"-- the BBCode field {repr(self._bbcode_field_name)} does not use attachments"
                )

        if self._main_attach_target_name is not None:
            target = getattr(opts.model, self._main_attach_target_name, None)
            if target is not None:
                target = target.field

            if not isinstance(target, models.ForeignKey) or not issubclass(target.remote_field.model, Attachment):
                raise ValueError(
                    f"Attachment target {repr(self._main_attach_target_name)} is not an FK to an Attachment."
                )

            if not target.null or not target.blank:
                raise ValueError(f"Attachment target {repr(self._main_attach_target_name)} is not nullable or blank.")

        if not self._has_attachments:
            return

        if "submit_attachments" in self.data:
            self.attachments_submitted = True
            try:
                attachments = self.files.getlist("attachment_upload[]")
            except AttributeError:  # We got a basic dict, not a MultiValueDict
                attachment = self.files.get("attachment_upload[]")
                attachments = [attachment] if attachment else []
            if not attachments:
                self.attachments_errors = [_("No file was submitted.")]

            for attachment in attachments:
                try:
                    Attachment.objects.upload_file(self._namespace, attachment)
                except ValidationError as err:
                    self.attachments_errors += err.messages

    def attachments_fieldset(self):
        """
        Displays the attachment selection/upload fieldset for BBCode
        attachments.
        """

        if not self._has_attachments:
            raise Exception("The BBCode field has no attachments")

        available_attach = Attachment.objects.get_available_attachments_for(self._namespace, self.instance.id or 0)

        ctx = {
            "static_url": settings.STATIC_URL,
            "available_attachments": available_attach,
            "attachments_errors": self.attachments_errors,
            "attachment_target_field": self._meta.bbcode_field,
        }

        # XXX filesizeformat passe un float à ngettext, qui génère un
        # warning (corrigé dans Django 3.0a1)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return render_to_string("form_widgets/attachments.html", ctx)

    def get_preview(self):
        """
        Get a preview for the current BBCode.
        If the BBCode parser raises errors, the errors are stored into the
        form (but only for the BBCode field -- other fields are not validated)
        and this function returns None.
        """

        if self.attachments_submitted:
            return None

        self._errors = ErrorDict()
        self.cleaned_data = {}

        field = self._bbcode_field
        try:
            value = field.widget.value_from_datadict(self.data, self.files, self.add_prefix(self._bbcode_field_name))
            value = field.clean(value)
            return value.as_html()
        except ValidationError as err:
            self.add_error(self._bbcode_field_name, err)
            return None
        finally:
            del self.cleaned_data

    def is_valid(self):
        if self.attachments_submitted:
            # Do not allow submitting the form when a file was uploaded…
            return False

        return super(BBCodeModelForm, self).is_valid()

    def full_clean(self):
        if self.attachments_submitted:
            # … but do not notify any field errors when it happens
            self._errors = ErrorDict()
            return

        super(BBCodeModelForm, self).full_clean()

    def save(self, commit=True):
        assert not self.attachments_submitted

        instance = super(BBCodeModelForm, self).save(commit=commit)

        if not self._has_attachments:
            return instance

        if self._main_attach_target_name is not None:
            value = self.cleaned_data[self._bbcode_field_name]
            main_attach = value.get_main_attachment()
            setattr(instance, self._main_attach_target_name, main_attach)

        if commit:
            self.commit_attachments()

        return instance

    def commit_attachments(self):
        if not self._has_attachments:
            return

        value = self.cleaned_data[self._bbcode_field_name]
        value.commit_attachments(self._namespace, self.instance.id)

    def _clean_fields(self):
        super(BBCodeModelForm, self)._clean_fields()

        if not self._has_attachments:
            return

        value = self.cleaned_data.get(self._bbcode_field_name, None)

        if value is not None:
            try:
                value.validate_attachments(self._namespace, self.instance.id)
            except ValidationError as err:
                self.add_error(self._bbcode_field_name, err)
